from auxi.simulation.io.writer import *
from openpyxl import Workbook
from auxi.simulation.path_engine import *
from auxi.simulation.investigation import *


class Writer_Excel(Writer):
    def __init__(self, file_path):
        Writer.__init__(self, file_path)

    def create_file_from_investigation(self, investigation):
        workbook = Workbook()
        sheet_ix = 0
        column_headers = ['Segment', 'Parameter', 'Description', 'Units',
                          'Value']

        for sc_n, sc_val in investigation.scenarios.items():
            row_ix = 1
            ws = workbook.create_sheet(title='Parameters(%s)' % sc_n,
                                       index=sheet_ix)
            for ix, col in enumerate(column_headers):
                ws.cell(row=row_ix, column=ix+1).value = col
            sheet_ix += 1
            for seg_n, seg_val in sc_val.segments.items():
                for p_n, p_val in seg_val.parameters.items():
                    row_ix += 1
                    default_val = str(p_val.default_value)
                    ws.cell(row=row_ix, column=1).value = seg_val.path
                    ws.cell(row=row_ix, column=2).value = p_val.path
                    ws.cell(row=row_ix, column=3).value = p_val.description
                    ws.cell(row=row_ix, column=4).value = p_val.units
                    ws.cell(row=row_ix, column=5).value = default_val

        #df.to_excel(file_path, sc_n, header=False, index=True)
        workbook.save(self.file_path)



from auxi.simulation.path_engine import *
from auxi.simulation.investigation import *


class Reader():
    def __init__(self, file_path):
        self.file_path = file_path

    def update_scenario_from_file(self, investigation, scenario_name):
        print("Abstract Method. Make use of a derived class.")



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 14 10:19:02 2015

@author: Christoff Kok, Thabisani Nigel Phuthi
"""
__version__ = "0.0.0"

from auxi.modelling.business import NamedObject
from auxi.simulation.scenario import *
from auxi.simulation.path_engine import *
from datetime import datetime


def create_investigation_from_auxi_object(auxi_object):
    parameter_path_dict = generate_parameter_paths(auxi_object)

    result = Investigation()
    scenario = result.scenarios["default"]

    path_dict = {}
    for p_path, p_prop in parameter_path_dict.items():
        path_dict[p_path[0]] = p_path[1]

    for seg_path, seg_obj in path_dict.items():
        scenario.add_segment(seg_obj, seg_path)

    for p_path, p_prop in parameter_path_dict.items():
        default_val = 0
        if type(p_prop) is str:
            default_val = ""
        elif type(p_prop) is datetime:
            default_val = datetime(2015, 1, 1, 0, 0, 0)
        scenario.segments[p_path[0]].create_parameter(str(p_path[2]), p_prop, "", default_val)
        # TODO: Append Results
        # TODO: Global Variables must be accessible via parameters.
        #       Calculation Engines as well.
    return result


class Investigation(NamedObject):
    """ Acts as a container for scenarios which store all the parameters of the
    techno-economic model.

    Parameters are time based values (can be made to vary according to date or a
    given reference time.
    Results are also time based values (Are stored according to the simulation's
    virtual date.)
    A submodel's parameters and results are stored under the submodel's
    respective 'segment'
    This division of parameters into segements ensure that there is no mix up of
    parameters between models.
    Furthermore, segments are grouped into scenarios, as the parameters can be
    changed according to the investigation scenario.
    The scenarios are the highest level of grouping in an investigation

    To add a scenario to an investigation use the command::
        investigation.add_scenario('Scenario name')
    """
    def __init__(self):
        NamedObject.__init__(self)
        self.scenarios = {}
        self.scenarios["default"] = Scenario("default")
        self.scenarios["default"].path = self.scenarios["default"].name

    def __getitem__(self, key):
        return self.scenarios[key]

    def add_scenario(self, name):
        self.scenarios[name] = Scenario(name)
        self.scenarios[name].path = name
        return self.scenarios[name]

    def get_parameter(self, scenario_name, segment_name, parameter_name):
        if segment_name in self.scenarios[scenario_name].segments.keys():
            segment = self.scenarios[scenario_name][segment_name]
        else:
            segment = self.scenarios["default"][segment_name]

        return segment[parameter_name]

    def get_result(self, scenario_name, segment_name, result_name):
        if segment_name in self.scenarios[scenario_name].segments.keys():
            segment = self.scenarios[scenario_name][segment_name]
        else:
            segment = self.scenarios["default"][segment_name]

        return segment[result_name]

    def _get_name_path(self, parameter_name):
        for scenario in self.scenarios:
            for segment in scenario.segments:
                for parameter in segment.parameters:
                    if parameter.name == parameter_name:
                        name_path = parameter.path
                    else:
                        print("Parameter does not exist")

        return name_path

    def run_scenario(self, io_reader, scenario_name):
        # Set models' parameters to that of the investigation io reader's file.
        io_reader.update_scenario_from_file(self, scenario_name)
        for scenario in self.scenarios:
            if scenario == scenario_name:
                self.scenarios[scenario].run()
                break



"""auxi.modelling.accounting.stock Package

The purpose of this package is to provide common classes and functions for the rest of the auxi.modelling.accounting.stock namespace"""

__version__ = "0.1.0"

from auxi.modelling.accounting.stock._stock import *


"""auxi for Python Package

The purpose of this package is to make auxi's modeling and simulation framework
available through a series of Python modules for easy access and use.

Python version:  3.4.0
auxi version: 0.1.0"""

__version__ = "0.1.0"



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 14 08:30:29 2015

@author: Christoff Kok, Thabisani Nigel Phuthi
"""
__version__ = "0.0.0"

from auxi.modelling.business import NamedObject
from auxi.simulation.path_engine import *
from auxi.modelling.business import TimeInterval


class Parameter(NamedObject):
    def __init__(self, name, source, units, default_value, description=""):
        NamedObject.__init__(self)
        self.name = str(name)
        if description is not None:
            self.description = description
        self._source = source
        self.units = units
        self.default_value = default_value
        self._path = None

        self.value_dict_intervals = {}
        self.value_dict_dates = {}

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        self._path = value

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        self._source = value

    def get_dt_diff(self, date_time1, date_time2, interval_type, interval_size):
        if interval_type == TimeInterval.Millisecond:
            return (date_time1.millisecond - date_time2.millisecond) * interval_size
        if interval_type == TimeInterval.Second:
            return (date_time1.second - date_time2.second) * interval_size
        if interval_type == TimeInterval.Minute:
            return (date_time1.minute - date_time2.minute) * interval_size
        if interval_type == TimeInterval.Hour:
            return (date_time1.hour - date_time2.hour) * interval_size
        if interval_type == TimeInterval.Day:
            return (date_time1.day - date_time2.day) * interval_size
        if interval_type == TimeInterval.Week:
            return (date_time1.day - date_time2.day) * 7 * interval_size
        if interval_type == TimeInterval.Month:
            return (date_time1.month - date_time2.month) * interval_size
        if interval_type == TimeInterval.Year:
            return (date_time1.year - date_time2.year) * interval_size

    def get_nearest_date_value(self, date):
        if date in self.value_dict_dates:
                return date, self.value_dict_dates[date]
        else:
            # return the previous set value
            prev_key = datetime.min
            found = False
            for key in sorted(self.value_dict_dates.keys()):
                if key > date:
                    break
                prev_key = key
                found = True
            if not found:
                return prev_key, self.default_value
            else:
                return prev_key, self.value_dict_dates[prev_key]

    def get_nearest_interval_value(self, interval):
        if interval in self.value_dict_intervals:
            return interval, self.value_dict_intervals[interval]
        else:
            # return the previous set value
            prev_key = 0
            found = False
            for key in sorted(self.value_dict_intervals.keys()):
                if key > interval:
                    break
                prev_key = key
                found = True
            if not found:
                return prev_key, self.default_value
            else:
                return prev_key, self.value_dict_intervals[prev_key]

    def get_nearest_value(self, interval, date_time,
                          interval_type, interval_size):
        if interval is None:
            if date_time is None:
                return self.default_value
            else:
                return self.get_nearest_date_value(date_time)[1]
        else:
            if date_time is None:
                return self.get_nearest_interval_value(interval)[1]
            else:
                int_ix, interval_val = self.get_nearest_interval_value(
                    interval)
                dt_ix, date_val = self.get_nearest_date_value(date_time)
                if dt_ix is None:
                    if int_ix is None:
                        return self.default_value
                    else:
                        return interval_val
                elif int_ix is None:
                    return date_val
                elif self.get_dt_diff(date_time, dt_ix, interval_type,
                                      interval_size) > int_ix * interval_size:
                    return date_val
                else:
                    return interval_val

    def value(self, date_query=None, interval=None, date_time=None,
              interval_type=None, interval_size=None, scenario=None):
        if date_query is None:
            result = self.get_nearest_value(interval, date_time, interval_type,
                                            interval_size)
        elif isinstance(date_query, datetime):
            result = self.get_nearest_value(None, date_query, interval_type,
                                            interval_size)
        else:
            try:
                val_ix = eval(str(date_query))
                if isinstance(val_ix, int):
                    result = self.get_nearest_value(val_ix, None, interval_type,
                                                    interval_size)
                else:
                    result = self.default_value
            except:
                pass

        if scenario is not None and isinstance(result, str):
            param_start_ix = result.rindex(".")
            seg_path = result[0:param_start_ix]
            seg_name = seg_path[result.index("::")+2:]
            param_path = result[param_start_ix+1:]

            if seg_name in scenario.segments:
                seg = scenario.segments[seg_name]
                if param_path in seg.parameters:
                    result = seg.parameters[param_path]._source
        return result

    def apply_value(self, seg_source, interval=None, date_time=None,
                    interval_type=None, interval_size=None, scenario=None):
        val = self.value(
            interval=interval,
            date_time=date_time,
            interval_type=interval_type,
            interval_size=interval_size,
            scenario=scenario)
        if isinstance(self._source, int):
            val = int(val)
        set_instance_value(seg_source, self.name, val)



__version__ = "0.1.0"

from auxi.core import NamedObject
from auxi.modelling.business import TimeBasedModel
from auxi.modelling.business import Clock
from auxi.modelling.business import Entity
from auxi.modelling.business import Component
from auxi.modelling.business import Activity
from auxi.modelling.business import BasicActivity
from auxi.modelling.business import CapitalLoanActivity
from auxi.modelling.business import CustomPythonActivity
from auxi.modelling.business import TimeInterval
from auxi.simulation.custom_python_object import *


NamedObject.is_param = lambda self, p: False

TimeBasedModel.is_param = lambda self, p: p in ['entityList',
                                                'totalIntervalsToRun',
                                                'clock']

Clock.is_param = lambda self, p: p in ['timeStepIntervalCount',
                                       'timeStepInterval',
                                       'startDateTime']

Entity.is_param = lambda self, p: p in ['componentList',
                                        'generalLedger',
                                        'taxRuleSet',
                                        'variableGroupList']

Component.is_param = lambda self, p: p in ['activityList',
                                           'componentList',
                                           'variableGroupList']

Activity.is_param = lambda self, p: p in ['executeInterval',
                                          'executionEndAtInterval',
                                          'executionStartAtInterval']

BasicActivity.is_param = lambda self, p: p in ['executeInterval',
                                               'executionEndAtInterval',
                                               'executionStartAtInterval',
                                               'amount',
                                               'date']

CapitalLoanActivity.is_param = lambda self, p: p in ['executeInterval',
                                                     'executionEndAtInterval',
                                                     'executionStartAtInterval',
                                                     'loanAmount',
                                                     'date',
                                                     'interestRate',
                                                     'periodInMonths',
                                                     'amountLeft',
                                                     'monthsLeft',
                                                     'monthlyPayment',
                                                     'currentInterestAmount']

CustomPythonActivity.is_param = lambda self, p: p in ['custom_python_object']


def run_Entity(self, segment):
    self.prepare_to_run()


def prepare_to_run_TimebasedModel(self, scenario):
    scenario.apply_parameter_values(0,
                                    self.clock.startDateTime,
                                    self.clock.timeStepInterval,
                                    1)
    self.prepare_to_run()


TimeBasedModel.prepare_to_run_from_scenario = prepare_to_run_TimebasedModel


def run_TimebasedModel(self, scenario):
    self.prepare_to_run_from_scenario(scenario)
    for ix in range(0, self.totalIntervalsToRun):
        # Apply the scenario's parameter values for this iteration before
        #   executing this month.
        scenario.apply_parameter_values(ix,
                                        self.clock.getDateTimeAtInterval(ix),
                                        self.clock.timeStepInterval,
                                        1)

        for entity in self.entityList:
            entity.run(self.clock, ix, self.currency)

TimeBasedModel.run_from_scenario = run_TimebasedModel
